﻿/*
 * Copyright (c) 2010, Koji Tsukada<tsuka@mobiquitous.com> 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 *    
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using System.IO;

namespace MobiServer.Devices.Xbee
{

    public partial class ControlDevice : MobiServer.ControlBase
    {
        #region Variables
        private CheckBox[] checkBoxAnalogs;
        private CheckBox[] checkBoxDigitals;
     
        private NumericUpDown[] numericAnalogs;
        private NumericUpDown[] numericAnalogTs;

        const int ANALOG_LENGTH = 4;
        const int DIGITAL_LENGTH = 13;

        string address = "";

        long serialNumber;
        long networkAddress;
        ControlPanel.AddressModes addressMode = ControlPanel.AddressModes.SerialNumber;
        bool updateStateFlag = false;

        public XbeeEventHandler xbeeEventHandler = null;

        MobiServer.TCPServerControl serverControl = null;
        #endregion

        #region Constructor / Initialization
        public ControlDevice()
        {
            InitializeComponent();
        }

        public ControlDevice(MobiServer.MessageControl message, MobiServer.TCPServerControl server, DataDevice device, ControlPanel.AddressModes mode)
            : base(message)
        {
            InitializeComponent();
      
            InitArray();
            InitState(device);

            this.serverControl = server;
            //control.serverWriteHandler += new MobiServer.WriteHandler(serverControl.OnServerWriteEvent);
            serverControl.AddReadHandler(new MobiServer.ServerEventHandler(OnServerReadEvent));

            this.addressMode = mode;
            this.serialNumber = device.SerialNumber;
            this.networkAddress = device.NetworkAddress;

            SetAddressMode(mode);   //device.SerialNumber, device.NetworkAddress
            
            xbeeEventHandler += new MobiServer.Devices.Xbee.XbeeEventHandler(OnXbeeEvent);
        }

        private void ControlDevice_Load(object sender, EventArgs e)
        {
            LoadSetting();
            SetTitle(serialNumber.ToString("X") + ":" + networkAddress.ToString("X"));
        }


        void InitArray()
        {
            numericAnalogs = new System.Windows.Forms.NumericUpDown[ANALOG_LENGTH];
            numericAnalogs[0] = numericUpDownAnalog1;
            numericAnalogs[1] = numericUpDownAnalog2;
            numericAnalogs[2] = numericUpDownAnalog3;
            numericAnalogs[3] = numericUpDownAnalog4;

            numericAnalogTs = new System.Windows.Forms.NumericUpDown[ANALOG_LENGTH];
            numericAnalogTs[0] = numericUpDownAnalogT1;
            numericAnalogTs[1] = numericUpDownAnalogT2;
            numericAnalogTs[2] = numericUpDownAnalogT3;
            numericAnalogTs[3] = numericUpDownAnalogT4;

            checkBoxAnalogs = new System.Windows.Forms.CheckBox[ANALOG_LENGTH];
            checkBoxAnalogs[0] = checkBoxAnalog1;
            checkBoxAnalogs[1] = checkBoxAnalog2;
            checkBoxAnalogs[2] = checkBoxAnalog3;
            checkBoxAnalogs[3] = checkBoxAnalog4;

            checkBoxDigitals = new System.Windows.Forms.CheckBox[DIGITAL_LENGTH];
            checkBoxDigitals[0] = checkBoxDigital1;
            checkBoxDigitals[1] = checkBoxDigital2;
            checkBoxDigitals[2] = checkBoxDigital3;
            checkBoxDigitals[3] = checkBoxDigital4;
            checkBoxDigitals[4] = checkBoxDigital5;
            checkBoxDigitals[5] = checkBoxDigital6;
            checkBoxDigitals[6] = checkBoxDigital7;
            checkBoxDigitals[7] = checkBoxDigital8;
            checkBoxDigitals[8] = checkBoxDigital9;
            checkBoxDigitals[9] = checkBoxDigital10;
            checkBoxDigitals[10] = checkBoxDigital11;
            checkBoxDigitals[11] = checkBoxDigital12;
            checkBoxDigitals[12] = checkBoxDigital13;

            return;
        }
      
        void InitState(DataDevice device)
        {
            for (int i = 0; i < ANALOG_LENGTH; i++)
            {
                numericAnalogs[i].Enabled = numericAnalogTs[i].Enabled = checkBoxAnalogs[i].Enabled = device.AnalogEnables[i];
                if (device.AnalogData != null && device.AnalogEnables[i])
                {
                    numericAnalogs[i].Value = (decimal)device.AnalogData[i];
                }
            }

            for (int i = 0; i < DIGITAL_LENGTH; i++)
            {
                checkBoxDigitals[i].Enabled = device.DigitalEnables[i];
                if (device.DigitalData != null && device.DigitalEnables[i])
                {
                    checkBoxDigitals[i].Checked = device.DigitalData[i];
                }
            }
        }

        #endregion

        #region Xbee Events

        public void OnXbeeEvent(object sender, XbeeEventArgs e)
        {
            if (CheckAddress(e.Device.SerialNumber, e.Device.NetworkAddress) && GetEnabled())
            {
                if (updateStateFlag)
                {
                    UpdateState(e.Device);
                }
                UpdateData(e.Device);
            }
        }

        public void UpdateState(DataDevice device)
        {
            for (int i = 0; i < ANALOG_LENGTH; i++)
            {
                if (numericAnalogs[i].Enabled != device.AnalogEnables[i])
                {
                    numericAnalogs[i].Enabled = numericAnalogTs[i].Enabled = checkBoxAnalogs[i].Enabled = device.AnalogEnables[i];
                }
            }

            for (int i = 0; i < DIGITAL_LENGTH; i++)
            {
                if (checkBoxDigitals[i].Enabled != device.DigitalEnables[i])
                {
                    checkBoxDigitals[i].Enabled = device.DigitalEnables[i];
                }
            }
        }

        public void UpdateData(DataDevice device)
        {
            if (device.AnalogData != null)
            {
                for (int i = 0; i < ANALOG_LENGTH; i++)
                {
                    if (numericAnalogs[i].Enabled &&
                          Math.Abs((int)device.AnalogData[i] - (int)numericAnalogs[i].Value) > (int)numericAnalogTs[i].Value
                        )
                    {
                        numericAnalogs[i].Value = (decimal)device.AnalogData[i];
                        numericAnalogs[i].Update();
                        checkBoxAnalogs[i].Checked = true;
                        DataServer server = new DataServer(DataServer.States.In, address, DataServer.Types.Analog, i, (int)device.AnalogData[i]);
                        CreateServerEvent(server);

                    }
                    else
                    {
                        if(checkBoxAnalogs[i].Checked)
                            checkBoxAnalogs[i].Checked = false;
                    }
                }
            }

            if (device.DigitalData != null ){
                for (int i = 0; i < DIGITAL_LENGTH; i++)
                {
                    if (checkBoxDigitals[i].Enabled)
                    {
                        if (checkBoxDigitals[i].Checked != device.DigitalData[i])
                        {
                            checkBoxDigitals[i].Checked = device.DigitalData[i];
                            checkBoxDigitals[i].Update();

                            DataServer server = new DataServer(DataServer.States.In, address, DataServer.Types.Digital, i, device.DigitalData[i]);
                            CreateServerEvent(server);
                        }
                    }
                }

            }
            Application.DoEvents();
        }

        void CreateServerEvent(DataServer server)
        {
            if (serverControl != null)
            {
                serverControl.WriteAllSockets(server.ToArray());
            }

        }


        //Not worked yet.
        public void OnServerReadEvent(object sender, MobiServer.ServerEventArgs e)
        {
            /*
            try
            {
                if (CheckDeviceType(e.Args))
                {
                   
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
            */
        }


        #endregion

        #region Address / Attach&Detach

        public void SetAddressMode(ControlPanel.AddressModes mode){

            this.addressMode = mode;
            switch (addressMode)
            {
                case ControlPanel.AddressModes.SerialNumber:
                    this.address = serialNumber.ToString("X");
                    break;
                case ControlPanel.AddressModes.NetworkAddress:
                    this.address = networkAddress.ToString("X");
                    break;
                //case ControlPanel.AddressModes.Both:
                //    this.address = serialNumber.ToString("X") + ":" + networkAddress.ToString("X");
                //    break;
            }
            textBoxAddress.Text = this.address;

            DataServer server = new DataServer(DataServer.States.Attach, this.address);
            CreateServerEvent(server);
        }

        public void SetUpdateStateFlag(bool flag)
        {
            this.updateStateFlag = flag;
        }

        public bool CheckAddress(long serialNumer, short networkAddress)
        {
            string address = "";
            switch (addressMode)
            {
                case ControlPanel.AddressModes.SerialNumber:
                    address = serialNumer.ToString("X");
                    break;
                case ControlPanel.AddressModes.NetworkAddress:
                    address = networkAddress.ToString("X");
                    break;
                //case ControlPanel.AddressModes.Both:
                //    address = serialNumer.ToString("X") + ":" + networkAddress.ToString("X");
                //    break;
            }

            return CheckAddress(address);
          
        }

        public string Address
        {
            get { return this.address;   }
        }

        public bool CheckAddress(string address)
        {
            Match m = Regex.Match(this.address, address, RegexOptions.IgnoreCase);
            if (m.Success)
            {

                return true;
            }
            else
            {
                return false;
            }
        }

        public bool GetEnabled()
        {
            return groupBoxBase.Enabled;
        }

        public void SetEnabled(bool flag)
        {
            if (!flag)
            {
                buttonDettach.Image = MobiServer.Devices.Xbee.Properties.Resources.attach;
                this.groupBoxBase.Enabled = false;
                //this.SaveSettingDevice();
            }
            else
            {
                this.groupBoxBase.Enabled = true;
                buttonDettach.Image = MobiServer.Devices.Xbee.Properties.Resources.close;
            }  
        }


        private void buttonDettach_Click(object sender, EventArgs e)
        {
            SetEnabled(!this.groupBoxBase.Enabled);
            if (groupBoxBase.Enabled)
            {
                CreateServerEvent(new DataServer(DataServer.States.Attach, this.address));
            }
            else
            {
                CreateServerEvent(new DataServer(DataServer.States.Dettach, this.address));
            }
        }

        #endregion


        #region Save/Load

        string GetSettingFile()
        {
            string dir = Application.StartupPath + @"\" + SETTING_DIRECTORY;

            if (!Directory.Exists(dir))
            {
                Directory.CreateDirectory(dir);
            }

            string filename = String.Format(@"{0}\{1}_{2}.txt", dir, DataServer.Header,serialNumber);

            return filename;
        }


        public void SaveSetting()
        {
            StreamWriter fileStream = null;
            string file = GetSettingFile();

            try
            {
                fileStream = new StreamWriter(file, false);

                for (int i = 0; i < numericAnalogTs.Length; i++)
                {
                    fileStream.Write("{0}\t", numericAnalogTs[i].Value);
                }
                fileStream.WriteLine();

                //for (int i = 0; i < checkBoxDigitals.Length; i++)
                //{
                //    fileStream.Write("{0}\t", checkBoxDigitals[i].Checked);
                //}
                //fileStream.WriteLine();
            }
            catch (IOException ex)
            {
                PrintError(ex.Message);
            }
            catch (Exception ex)
            {
                PrintError("Saving " + Path.GetFileName(file));
            }
            finally
            {
                if (fileStream != null)
                {
                    fileStream.Close();
                    fileStream = null;
                }
            }
        }

        public void LoadSetting()
        {
            StreamReader fileStream = null;
            string file = GetSettingFile();

            try
            {
                if (File.Exists(file))
                {

                    fileStream = new StreamReader(file);

                    string line = null;
                    string[] args = null;

                    line = fileStream.ReadLine();
                    args = line.Split("\t".ToCharArray());
                    for (int i = 0; i < numericAnalogTs.Length; i++)
                    {
                        numericAnalogTs[i].Value = int.Parse(args[i]);
                    }

                    //line = fileStream.ReadLine();
                    //args = line.Split("\t".ToCharArray());
                    //for (int i = 0; i < checkBoxDigitals.Length; i++)
                    //{
                     //   checkBoxDigitals[i].Checked = bool.Parse(args[i]);
                    //}
                }

            }
            catch (IOException ex)
            {
                PrintMessage(ex.Message);
            }
            catch (Exception ex)
            {
                PrintError("Loading " + Path.GetFileName(file));

            }
            finally
            {
                if (fileStream != null)
                {
                    fileStream.Close();
                    fileStream = null;
                }
            }
        }


        #endregion
    }
}
